package com.sourcetohtml;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

/**
 * factory of FileDetailGenerator
 * Current factory will hold a mapping between
 * file extensions and FileDetailGenerator implementations
 * The mapping is read from configuration file
 * @author jcelak
 */
public class FileDetailGeneratorFactory {

	private static FileDetailGeneratorFactory singleton = new FileDetailGeneratorFactory();
	
	/** generator used for directories */
	private FileDetailGenerator dirGen;
	
	/** generator used if no mapping is defined*/
	private FileDetailGenerator unknownGen;
	
	/** class name -> singelton instance*/
	private Map<String,FileDetailGenerator> allGenerators = new HashMap<String, FileDetailGenerator>();
		
	private FileDetailGeneratorFactory()
	{
		dirGen = getInstance(Config.getInstance().getProperty("directoryGenerator", 
				DirectoryDetailGenerator.class.getName()));
		unknownGen = getInstance(Config.getInstance().getProperty("defaultGenerator", 
				UnknownGenerator.class.getName()));
	}
	
	private synchronized FileDetailGenerator getInstance(String clazz)
	{
		FileDetailGenerator result = allGenerators.get(clazz);
		if (result == null)
		{
			try {
				Class c = Class.forName(clazz);
				result = (FileDetailGenerator)c.newInstance();
				allGenerators.put(clazz,result);
			} catch (Exception e) {
				throw new RuntimeException(e.getMessage(),e);
			} 
		}
		return result;
	}
	
	public static FileDetailGeneratorFactory getInstance()
	{
		return singleton;		
	}
	
	/**
	 * returns a generator for passed file
	 * @param f input file
	 * @return file HTML generator that will be used
	 */
	public FileDetailGenerator getGenerator(File f)
	{
		if (f.isDirectory())
			return dirGen;
		else
		{		
			return getInstance(
					Config.getInstance().getProperty("fileMapping." + getFileExtension(f), 
					unknownGen.getClass().getName()));
		}		
	}
	
	private String getFileExtension(File f)
	{
		String name = f.getName();
		int index = name.lastIndexOf('.');
		if (index >= 0 && index<name.length())
		{
			return name.substring(index+1).toLowerCase();
		}
		return name.toLowerCase();
	}
}
